#!/usr/bin/python

# Copyright (C) 2012-2013 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Lightweight command line client for linknx. It is aimed at reading or writing object values from/to linknx.
When reading objects, this script outputs a line per object, each line composed of the object id and its corresponding value separated with spaces.

SYNTAX:
		pyknxclient.py [-h host] [-p port] [-v level] -r id1 [-r id2 [...]]
		pyknxclient.py [-h host] [-p port] [-v level] -R pattern
		pyknxclient.py [-h host] [-p port] [-v level] -w id value

		Where id, id1, id2, ... are object identifiers specified in the linknx configuration XML with the 'id' attribute.

OPTIONS:
		-h, --host      Hostname of the machine running the linknx daemon (default is 'localhost').
		-p, --port      Port linknx listens on (default is 1028).
		-v, --verbose   Level of verbosity. Value must be one of the logging module (error, warning, info, debug)
		-r, --read      Read value of object with given id. Can occur multiple times with various identifiers.
		-R, --regex     Read value of all objects whose identifiers match the given regex pattern. The pattern must comply with the 're' python module.
		-w, --write     Writes a new value to the object of given identifier.
	        --help      Display this help message and quit.

EXAMPLES:
		Read all objects:
			pyknxclient.py -R ".*"

		Read two objects:
			pyknxclient.py -r KitchenLights -r LivingRoomLights

		Turn off kitchen lights:
			pyknxclient.py -w KitchenLights off
			pyknxclient.py -w KitchenLights false
			pyknxclient.py -w KitchenLights 0
"""

import sys
import getopt
import traceback
import logging
from xml.dom.minidom import parseString
from threading import *
from pyknx import logger
from pyknx.linknx import *

def printUsage():
	print __doc__

if __name__ == '__main__':
	logger.initLogger(None, logging.INFO, usesDetailedLogging=False)

	try:
		options, remainder = getopt.getopt(sys.argv[1:], 'r:w:h:p:v:R:', ['read=', 'write=', 'regex=', 'host=', 'port=','verbose=','help'])
	except getopt.GetoptError as err:
		logger.reportError(sys.exc_info()[1])
		sys.exit(2)

	# Parse command line arguments.
	reads = False
	writes = False
	objectIds = []
	isRegex = False
	host = 'localhost'
	port = 1028
	verbosity = logging.WARNING
	for option, value in options:
		if option == '-r' or option == '--read':
			reads = True
			objectIds.append(value)
		elif option == '-R' or option == '--regex':
			reads = True
			objectIds.append(value)
			isRegex = True
		elif option == '-w' or option == '--write':
			writes = True
			objectIds.append(value)
		elif option == '-h' or option == '--host':
			host = value
		elif option == '-p' or option == '--port':
			port = value
		elif option == '-v' or option == '--verbose':
			verbosity = logger.parseLevel(value)
		elif option == '--help':
			printUsage()
			sys.exit(1)
		else:
			print 'Unrecognized option ' + option
			sys.exit(2)

	logger.initLogger(None, verbosity, usesDetailedLogging=False)

	if reads == writes:
		logger.reportError('Expecting -r|--read or -w|--write and not both.')
		sys.exit(2)

	if writes and len(objectIds) > 1:
		logger.reportError('Can only write one object.')
		sys.exit(2)

	valueToWrite = None
	if writes:
		if not remainder:
			print 'No value specified.'
			printUsage()
			sys.exit(3)
		valueToWrite = remainder[0]
		del remainder[0]

	if remainder:
		logger.reportError('Too many arguments: ' + str(remainder))
		sys.exit(4)

	# Start linknx.
	linknx = Linknx(host, int(port))
	try:
		if reads:
			report={}
			if not isRegex:
				for objId in objectIds:
					report[objId] = linknx.getObject(objId).value
			else:
				for obj in linknx.getObjects(objectIds[0]):
					report[obj.id] = obj.value

			# No object.
			if not report:
				logger.reportWarning('No object of given id.')
				sys.exit(10)

			# Count tabs to align columns.
			longestId = max([len(id) for id in report.keys()])
			for o, v in report.iteritems():
				spaceCount = longestId - len(o)
				spaces=''
				while spaceCount > 0:
					spaces+=' '
					spaceCount -= 1
				print '{0} {2} {1}'.format(o, v, spaces)
		elif writes:
			linknx.getObject(objectIds[0]).value = valueToWrite
	except Exception as e:
		if verbosity == logging.DEBUG:
			logger.reportException()
		else:
			logger.reportError(sys.exc_info()[1])
		sys.exit(3)



#!/usr/bin/python

# Copyright (C) 2012-2013 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

import sys
import re
import getopt
import time
from xml.dom.minidom import parseString
from threading import *
from pyknx import tcpsocket, logger

class Linknx:

	"""
	The wrapper of an instance of Linknx.

	This class eases access to Linknx functionalities: it can retrieve objects (see getObject() and getObjects()), or configuration.
	"""

	class InvalidObjectIdException(Exception):
		""" The object id is not valid in the Linknx instance. """
		def __init__(self, objectId):
			self._objectId = objectId

		def __str__(self):
			return 'Object ' + self._objectId + ' does not exist.'

	def __init__(self, hostname='localhost', port=1028):
		""" Initialize a Linknx wrapper. """
		self._host = hostname
		self._port = port
		self._config = None
		self._objectConfig = None
		self._objects = {}

	@property
	def host(self):
		return self._host

	@property
	def port(self):
		return self._port

	@property
	def config(self):
		"""
		Get the raw XML configuration currently used by the running Linknx instance.

		See objectConfig property for an alternative of these raw data.

		"""

		if self._config is None:
			xmlConfig = self.sendMessage("<read><config></config></read>").getElementsByTagName('read')[0]
			self._config = xmlConfig.getElementsByTagName('config')[0]

		return self._config

	@property
	def objectConfig(self):
		""" Return a dictionary of the configuration of each Linknx object. Keys are object ids, values are ObjectConfig instances. """
		if not self._objectConfig:
			objectsConfigNode = self.config.getElementsByTagName('objects')[0]
			self._objectConfig = {} # key is objectId, value is an ObjectConfig
			for objectConfigNode in objectsConfigNode.getElementsByTagName('object'):
				objectConfig = ObjectConfig(objectConfigNode)
				self._objectConfig[objectConfig.id] = ObjectConfig(objectConfigNode)

		return self._objectConfig

	def waitForRemoteConnectionReady(self):
		"""
		Wait for Linknx's XML server to accept incoming connections.

		This method should be called if unsure about when the Linknx has been started. It may take some time to get ready.
		This method attempts to connect to Linknx during 10 seconds and raises an Exception if Linknx still is unreachable after this delay.

		"""
		# Ask for config.
		logger.reportInfo('Start connecting to linknx.')
		attemptId = 0
		maxAttemptCount = 10
		while attemptId < maxAttemptCount:
			attemptId += 1
			try:
				conf = self.config

				# Linknx is ready if we reach this point.
				logger.reportInfo('Linknx is up and ready, let\'s start.')
				return
			except Exception as e:
				logger.reportInfo('Linknx is not yet ready...  (attempt {0}/{1})'.format(attemptId, maxAttemptCount))
				time.sleep(1)

		raise Exception('Linknx is not reachable.')

	def getObject(self, id):
		""" Get the object of given identifier. """
		obj = self._objects.get(id)
		if obj is None:
			obj = Object(id, self)
			self._objects[id] = obj
		return obj

	def getObjects(self, pattern='.*'):
		""" Get the objects whose identifiers match the given regex pattern. Default pattern matches all identifiers, thus calling getObjects() returns all objects. """
		regex = re.compile(pattern)
		objects = []
		for id in self.objectConfig.iterkeys():
			if regex.search(id) is None: continue
			objects.append(self.getObject(id))
		return objects

	def _getErrorFromXML(self, answer):
		errorMessage = ""
		for child in answer.childNodes:
			if child.nodeType == child.TEXT_NODE:
				errorMessage += child.data
		return errorMessage

	def sendMessage(self, message):
		"""
		Sends an XML message to Linknx.

		This function is implemented mainly for internal purposes. The end user is unlikely to call it directly.
		message -- An XML request that follows Linknx XML protocol.
		Returns an XML document that corresponds to Linknx answer.

		"""
		# logger.reportDebug('Sending message to linknx: ' + message)
		s = tcpsocket.Socket()
		s.connect((self._host, self._port))
		try:
			answer = s.sendData(message)
			# logger.reportDebug('Linknx answered ' + answer)
			return parseString(answer[0:answer.rfind(chr(4))])
		finally:
			s.close()

class ObjectConfig:
	def __init__(self, configNode):
		"""	Configuration of an object in Linknx. """
		self.id = configNode.getAttribute('id')
		self.xml = configNode
		self.type = configNode.getAttribute('type')
		self.gad = configNode.getAttribute('gad')
		self.callback = configNode.getAttribute('pyknxcallback')
		self.init = configNode.getAttribute('init')
		firstTypeDigit = self.type[0:self.type.find('.')]
		if firstTypeDigit == '1':
			self.typeCategory = 'bool'
		elif firstTypeDigit in ['5', '6', '7', '8', '9', '12', '13']:
			if self.type in ['5.001', '5.003', '9.xxx']:
				self.typeCategory='float'
			else:
				self.typeCategory = 'int'
		elif firstTypeDigit == '16':
			self.typeCategory='string'
		else:
			logger.reportWarning('Object ' + self.id + ' has an unsupported type ' + self.type)
			self.typeCategory = 'unknown'

	# @property
	# def defaultValue(self):
		# """ Return the default value of the object. """
		# if self.type == 'bool':
			# return False
		# elif self.type == 'int':
			# return 0
		# elif self.type == 'float':
			# return 0.0
		# elif self.type == 'string':
			# return ''
		# else:
			# return None

class Object(object):

	""" Linknx object. """

	def __init__(self, id, linknx):
		"""
		Initialize an object from Linknx.

		id -- Identifier of the object. Corresponds to the id attribute in XML configuration.
		linknx -- Linknx instance that provides the object.
		"""
		self._id = id
		self._linknx = linknx
		if not linknx.objectConfig.has_key(id):
			raise Linknx.InvalidObjectIdException(id)

		self._objectConfig = linknx.objectConfig[id]

	@property
	def id(self):
		""" Object identifier. This is the id attribute in XML configuration. """
		return self._id

	@property
	def gad(self):
		""" Group Address of the object. This is the gad attribute in XML configuration. """
		return self._objectConfig.gad

	@property
	def linknx(self):
		""" Return the Linknx instance that provides the object. """
		return self._linknx

	@property
	def xml(self):
		""" Return the xml element corresponding to this object in Linknx configuration. """
		return self._objectConfig.xml

	@property
	def type(self):
		""" Return the string corresponding to the type attribute in XML configuration. """
		self._objectConfig.type

	@property
	def value(self):
		""" Read object's value from linknx. """
		message='<read><objects><object id="' + self._id + '"/></objects></read>'

		answerDom = self._linknx.sendMessage(message)

		readNodes = answerDom.getElementsByTagName("read")
		status = readNodes[0].getAttribute("status")
		valueStr = None
		if status == "success":
			objectValues = {}
			objectsNodes = readNodes[0].getElementsByTagName("objects")
			objectNodes = objectsNodes[0].getElementsByTagName("object")
			for objectNode in objectNodes:
				objectId = objectNode.getAttribute("id")
				objectValue = objectNode.getAttribute("value")
				objectValues[objectId] = objectValue
			valueStr = objectValues[self._id]
		else:
			raise Exception(self._linknx._getErrorFromXML(readNodes[0]))

		if self._objectConfig.typeCategory == 'bool':
			return valueStr in ['on', '1']
		elif self._objectConfig.typeCategory == 'int':
			return int(valueStr)
		else:
			return valueStr

	@value.setter
	def value(self, objValue):
		""" Write object's value to linknx. """
		# Convert value to the linknx format.
		logger.reportDebug('Attempting to set value of {0} to {1}'.format(self._id, objValue))
		if self._objectConfig.typeCategory == 'bool':
			if isinstance(objValue,bool):
				objectValue = 'on' if objValue else 'off'
			else:
				objValueStr = str(objValue).lower()
				if objValueStr in ['true', 'on', 'yes', '1']:
					objectValue = 'on'
				elif objValueStr in ['false', 'off', 'no', '0']:
					objectValue = 'off'
				else:
					raise Exception('For object {1}: Unable to convert {0} to boolean.'.format(objValue, self._id))
		elif self._objectConfig.typeCategory == 'int':
			objectValue = str(objValue)
		elif self._objectConfig.typeCategory == 'float':
			objectValue = str(objValue)
		elif self._objectConfig.typeCategory == 'string':
			objectValue = str(objValue)
		else:
			raise Exception('Unsupported type ' + self._objectConfig.typeCategory)

		if not objValue is objectValue:
			logger.reportDebug('Value has been converted to ' + str(objectValue))

		message='<write><object id="' + self._id + '" value="' + objectValue + '"/></write>'
		answerDom = self._linknx.sendMessage(message)
		writeNodes = answerDom.getElementsByTagName("write")
		status = writeNodes[0].getAttribute("status")
		if status != "success":
			raise Exception(self._linknx._getErrorFromXML(writeNodes[0]))

	def __repr__(self):
		return self.id

	def __str__(self):
		return self.id



#!/usr/bin/python

# Copyright (C) 2012-2013 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
USAGE:
	pyknxcommunicator.py [-l linknxaddress] [-c communicatoraddress] -f userfile

OPTIONS:
    -l --linknx-addr  Address of the linknx daemon. This must be of the form <hostname:port> of <ipaddress:port>. Default is localhost:1028
    -c --comm-addr    Address the communicator will bind to. This must be of the form <hostname:port> of <ipaddress:port>. Default is localhost:1029
    -f --user-file    Filename of the python script that implement the callbacks specified in linknx XML configuration.
       --verbose      Verbosity level, amongst [debug, info, warning, error]. Default is info.
       --help         Display this help message and exit.
"""

import sys
import getopt
import signal
import time
import logging
from pyknx import logger
from pyknx.linknx import *
from pyknx.communicator import *

def signal_handler(signal, frame):
	logger.reportInfo('Terminating...')
	communicator.stopListening()

def printUsage():
	print __doc__

def parseAddress(addrStr, option):
	ix = addrStr.find(':')
	if ix < 0:
		raise Exception('Malformed value for ' + option +'. Expecting a tuple (hostname:port)')
	return (addrStr[0:ix], int(addrStr[ix + 1:]))

if __name__ == '__main__':
	logger.initLogger(None, logging.INFO)
	try:
		try:
			options, remainder = getopt.getopt(sys.argv[1:], 'c:l:f:', ['comm-addr=', 'linknx-addr=', 'user-file=', 'log-file=', 'verbose=', 'pid-file=', 'help'])
		except getopt.GetoptError as err:
			logger.reportException()
			sys.exit(2)

		# Parse command line arguments.
		communicatorAddress = ('127.0.0.1',1029)
		linknxAddress = ('127.0.0.1',1028)
		userFile = None
		configFile = None
		verbosity = logging.INFO
		logFile = None
		pidFile = None
		# TODO Add support for script arguments (already implemented in
		# Communicator).
		for option, value in options:
			if option == '-c' or option == '--comm-addr':
				communicatorAddress = parseAddress(value, option)
			elif option == '-l' or option == '--linknx-addr':
				linknxAddress = parseAddress(value, option)
			elif option == '-f' or option == '--user-file':
				userFile = value
			elif option == '--help':
				printUsage()
				sys.exit(1)
			elif option == '--log-file':
				logFile = value
			elif option == '--pid-file':
				pidFile = value
			elif option == '--verbose':
				verbosity = logger.parseLevel(value)
			else:
				logger.reportError('Unrecognized option ' + option)
				sys.exit(2)

		# Init logger.
		if not logFile is None:
			logger.initLogger((logFile, verbosity), None)
		else:
			logger.initLogger(None, verbosity)

		if not userFile:
			printUsage()
			sys.exit(1)

		if remainder:
			logger.reportError('Unknown arguments {0}.'.format(remainder))
			printUsage()
			sys.exit(3)

		# Create pid file.
		# Start linknx.
		linknx = Linknx(linknxAddress[0], linknxAddress[1])

		# Start communicator.
		communicator = Communicator(linknx, userFile, communicatorAddress)
		communicator.startListening()

		signal.signal(signal.SIGINT, signal_handler)
		signal.signal(signal.SIGTERM, signal_handler)

		# Main loop.
		while communicator.isListening:
			time.sleep(1)
	except:
		logger.reportException()



#!/usr/bin/python

# Copyright (C) 2012-2013 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Modifies an XML config file for Linknx so that it allows for communication with an instance of pyknxcommunicator.py
This script adds an ioport and a rule for each object that holds a pyknxcallback attribute in the source XML.
It always starts by cleaning any autogenerated rule from the input file and generates new ones unless the --clean option is passed.

USAGE:
	pyknxconf.py [-c communicatoraddress] -i inputfile [-o outputfile]

OPTIONS:
    -c --comm-addr <host:port>   Address of the communicator. This argument must specify the hostname or the ip address followed by a colon and the port to listen on. Default is localhost:1029'
    -i --inputfile               Filename of the original linknx xml config file. This file is modified by this script if no output file is specified.'
    -o --outputfile              Filename of the modified linknx xml config file to write. If missing, this argument defaults to the same file than the one specified with --inputfile.'
    -p --rule-prefix             Prefix for the rules generated by this script. Default is "pyknxrule_"'
    -v,--verbose                 Level of verbosity. Value must be one of the logging module (error, warning, info, debug)
       --clean                   Clean rules that were generated by this script but do not generate new rules. Rules to delete are those whose id starts by the rule prefix'
       --help                    Display this help message and exit.'
"""

from pyknx.configurator import Configurator
import getopt
import sys
import logging
from pyknx import logger

def printUsage():
	print __doc__

def parseAddress(addrStr, option):
    ix = addrStr.find(':')
    if ix < 0:
        raise Exception('Malformed value for ' + option +'. Expecting a tuple (hostname:port)')
    return (addrStr[0:ix], addrStr[ix + 1:])

if __name__ == '__main__':
    try:
        options, remainder = getopt.getopt(sys.argv[1:], 'c:i:o:p:', ['comm-addr=', 'input-file=', 'output-file=', 'rule-prefix=', 'verbose', 'clean', 'help'])
    except getopt.GetoptError as err:
        print str(err)
        sys.exit(2)

    # Parse command line arguments.
    communicatorAddress = ('127.0.0.1',1029)
    inputFile = None
    outputFile = None
    rulePrefix = 'pyknxrule_'
    cleanOnly = False
    verbosity = logging.INFO
    for option, value in options:
        if option == '-c' or option == '--comm-addr':
            communicatorAddress = parseAddress(value, option)
        elif option == '-i' or option == '--input-file':
            inputFile = value
        elif option == '-o' or option == '--output-file':
            outputFile = value
        elif option == '-p' or option == '--rule-prefix':
            rulePrefix = value
        elif option == '--verbose':
            verbosity = logging.DEBUG
        elif option == '--clean':
            cleanOnly = True
        elif option == '--help':
            printUsage()
            sys.exit(1)
        else:
            print 'Unrecognized option ' + option
            sys.exit(2)

    if not inputFile:
        printUsage()
        sys.exit(1)

    if not outputFile:
        outputFile = inputFile

    # Configure logger.
    logger.initLogger(None, verbosity)

    # Start configurator.
    configurator = Configurator(inputFile, outputFile, communicatorAddress, rulePrefix)

    # Generate config.
    configurator.cleanConfig()
    if not cleanOnly:
        configurator.generateConfig()
    configurator.writeConfig()



#!/usr/bin/python

# Copyright (C) 2012-2013 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

from distutils.core import setup

setup(	name='pyknx',
		version='1.0.1',
		description='Python bindings for KNX',
		long_description=''.join(open('README').readlines()),
		author='Cyrille Defranoux',
		author_email='knx@aminate.net',
		maintainer='Cyrille Defranoux',
		maintainer_email='knx@aminate.net',
		license='GNU Public General License',
		url='https://pypi.python.org/pypi/pyknx/',
		packages=['pyknx'],
		data_files=['LICENSE'],
		scripts=['pyknxcommunicator.py', 'pyknxcall.py', 'pyknxclient.py', 'pyknxconf.py'])
